/*
 * Directory
 * 	version 1.0.0
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <cstdio>
#include <algorithm>
#include "DirectoryEntry.hpp"

using namespace std;

namespace DirectoryEntryPackage {
/* ================================================================
 * DirectoryNX
 * ================================================================ */

/* ----------------------------------------------------------------
 * Constructors / Destructors
 * ---------------------------------------------------------------- */

/* Constructor
 * <param>	Dir
 * 		Name
 * <exception>	EntryNOTFoundException
 *		ParameterNOTSpecifiedException
 *		IllegalParameterException
 */
Directory::Directory(const char * Dir, const char * Name)
	: DirectoryEntry(Dir, Name)
{
	if (!(Stat_.st_mode & S_IFDIR))
		throw IllegalParameterException("Directory::Directory(const char*,const char*)", "Name", Name, 0);
}

/* ----------------------------------------------------------------
 * Methods
 * ---------------------------------------------------------------- */

/* q */
void Directory::SearchChildren(bool recursive)
{
	string	dir = GetPathName();
	DIR*	dp;
	struct dirent*	entry;
	
	delete_children();
	
	/* fBNg̃t@Cꗗ𒊏o */
	if ((dp=opendir(dir.c_str())) == NULL)
		throw IllegalParameterException("Directory::SearchChildren()", "Dir", dir.c_str(), errno);
	while ((entry = readdir(dp)) != NULL){
		if (strcmp(entry->d_name, DIRSTR_CURRENT) != 0 && strcmp(entry->d_name, DIRSTR_PARENT) != 0) {
			DirectoryEntry*	found = Create(dir.c_str(), entry->d_name);
			if (found != NULL) {
				std::transform(		/* t@CɕϊĊi[ */
					entry->d_name,
					&entry->d_name[strlen(entry->d_name)],
					entry->d_name,
					(int (*)(int))std::tolower
				);
				SubEntries_.insert(DirectoryEntryCollection::value_type(entry->d_name, found));
			}
		}
	}
	closedir(dp);
	
	/* TuEfBNgċAIɌ */
	if (recursive) {
		for (DirectoryEntryCollection::iterator itr=SubEntries_.begin(); itr!=SubEntries_.end(); ++itr) {
			DirectoryEntry *	entry = dynamic_cast<DirectoryEntry*>(itr->second);
			if (!entry->IsTerminalNode())
				entry->SearchChildren(recursive);
		}
	}
	
	Searched_ = true;
}

/* g̕𐶐ĕԂ */
DirectoryEntry & Directory::Clone(void) const
{
	DirectoryEntry*	other = new Directory(Dir_.c_str(), Name_.c_str());
	*other = *this;
	return *other;
}

} /* namespace DirectoryEntryPackage */
/*-------1---------2---------3---------4---------5---------6---------7---------8---------9--------*/
/*  end of text */
