/*
 * DirectoryEntry
 * 	version 1.0.0
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/cygwin.h>
#include <cstdio>
#include <algorithm>
#include <iostream>
#include "DirectoryEntry.hpp"

using namespace std;

namespace DirectoryEntryPackage {
/* ================================================================
 * DirectoryEntryNX
 * ================================================================ */

/* ----------------------------------------------------------------
 * Statics
 * ---------------------------------------------------------------- */

/* fBNg(Dir)ăt@CEtH_𒊏o
 * <param>	Dir		fBNg
 * 		Entries		i[p
 *		recursive	true:TuEfBNgċAIɌ
 * <exception>	EntryNOTFoundException
 *		ParameterNOTSpecifiedException
 *		IllegalParameterException
 * <set>	Entries
 */
void DirectoryEntry::SearchChildren(const char * Dir, DirectoryEntryCollection & Entries, bool recursive)
{
	DIR*	dp;
	struct dirent*	entry;
	
	if (Dir == NULL)
		throw ParameterNOTSpecifiedException("DirectoryEntry::SearchChildren()", "Dir", "", 0);
	if (strcmp(Dir, "") == 0)
		throw IllegalParameterException("DirectoryEntry::SearchChildren()", "Dir", "", 0);
	
	/* fBNg̃t@Cꗗ𒊏o */
	if ((dp=opendir(Dir)) == NULL)
		throw IllegalParameterException("DirectoryEntry::SearchChildren()", "Dir", Dir, errno);
	while ((entry = readdir(dp)) != NULL){
		if (strcmp(entry->d_name, DIRSTR_CURRENT) != 0 && strcmp(entry->d_name, DIRSTR_PARENT) != 0) {
			DirectoryEntry*	found = Create(Dir, entry->d_name);
			if (found != NULL) {
				std::transform(		/* t@CɕϊĊi[ */
					entry->d_name,
					&entry->d_name[strlen(entry->d_name)],
					entry->d_name,
					(int (*)(int))std::tolower
				);
				Entries.insert(DirectoryEntryCollection::value_type(entry->d_name, found));
			}
		}
	}
	closedir(dp);
	
	/* TuEfBNgċAIɌ */
	if (recursive) {
		for (DirectoryEntryCollection::iterator itr=Entries.begin(); itr!=Entries.end(); ++itr) {
			DirectoryEntry *	entry = dynamic_cast<DirectoryEntry*>(itr->second);
			if (!entry->IsTerminalNode())
				entry->SearchChildren(recursive);
		}
	}
}

/* DirectoryEntryIuWFNg𐶐 */
DirectoryEntry* DirectoryEntry::Create(const char * Dir, const char * Name)
{
	struct stat	Stat;
	
	string	path = GetPathName(Dir, Name);
	if (!GetAttribute(path.c_str(), &Stat))
		return NULL;
	
	if (Stat.st_mode & S_IFDIR)
		return new Directory(Dir, Name);
	else	return new File(Dir, Name);
}

/* t@C̑擾 */
bool DirectoryEntry::GetAttribute(const char * Path, struct stat* Stat)
{
	if (stat(Path, Stat) == 0)
		return true;
	else	return false;
}

/* PATH(O+ꏊ)Ԃ */
string DirectoryEntry::GetPathName(const std::string & Dir, const std::string & Name)
{
	string	separator;
	if (Dir.find("/") != string::npos)
		separator = "/";
	else if (Dir.find("\\") != string::npos)
		separator = "\\";
	else	separator = "/";
	
	string	path;
	if (Dir[Dir.length() - 1] == separator[0])
		path = Dir + Name;
	else	path = Dir + separator + Name;
	
	return path;
}

/* ----------------------------------------------------------------
 * Constructors / Destructors
 * ---------------------------------------------------------------- */

/* Constructor
 * <param>	Dir
 * 		Entries
 * <exception>	IllegalParameterException
 *		ClassInitializeException
 */
DirectoryEntry::DirectoryEntry(const char * Dir, const char * Name)
	: Dir_(Dir), Name_(Name), SubEntries_(), Searched_(false), Stat_()
{
	if (Name_.length() == 0)
		throw IllegalParameterException("DirectoryEntry::DirectoryEntry(const char*,const char*)", "Name", Name, 0);
	if (!get_attribute())
		throw ClassInitializeException("DirectoryEntry::DirectoryEntry(const char*,const char*)", "Dir/Name", Dir_ + "/" + Name_, errno);
}

/* ----------------------------------------------------------------
 * Methods
 * ---------------------------------------------------------------- */

/* gq */
bool DirectoryEntry::HasExtension(const std::string & Ext) const
{
	if (Ext != ".") {
		string	ext = Ext;
		std::transform(ext.begin(), ext.end(), ext.begin(), (int (*)(int))std::tolower);
		string	name = Name_;
		std::transform(name.begin(), name.end(), name.begin(), (int (*)(int))std::tolower);
		string::size_type	pos = name.rfind(ext);
		if (pos != string::npos && name.substr(pos) == ext)
			return true;
		else	return false;
	} else {
		if (Name_.find(".") == string::npos)
			return true;
		else	return false;
	}
}

/* --------------------------------
 * Path
 * -------------------------------- */

/* PATH(O+ꏊ)Ԃ */
string DirectoryEntry::GetPathName(void) const
{
	return GetPathName(Dir_, Name_);
}

std::string DirectoryEntry::GetFullPathName(void) const
{
	string	path = GetPathName();
	char	fullpath[260];
	cygwin_conv_to_full_posix_path(path.c_str(), fullpath);
	return string(fullpath);
}

/* --------------------------------
 * t : ANZX
 * -------------------------------- */
//	const char *	wday[] = {"", "", "", "", "", "", "y"};

/* ŐṼANZXԂ */
long DirectoryEntry::GetAccessedTime(void) const
{
	return Stat_.st_atime;
}

/* ŐṼANZXԂ */
void DirectoryEntry::GetAccessedTime(struct tm * dst) const
{
	struct tm*	tmp = localtime(&Stat_.st_atime);
	*dst = *tmp;
}

/* ANZXt𐶐 */
void DirectoryEntry::GetAccessedDateString(char dst[]) const
{
	get_datestring(localtime(&Stat_.st_atime), dst);
}

string DirectoryEntry::GetAccessedDateString(void) const
{
	char	dst[16];
	get_datestring(localtime(&Stat_.st_atime), dst);
	return string(dst);
}

/* ANZX𐶐 */
void DirectoryEntry::GetAccessedTimeString(char dst[]) const
{
	get_timestring(localtime(&Stat_.st_atime), dst);
}

string DirectoryEntry::GetAccessedTimeString(void) const
{
	char	dst[16];
	get_timestring(localtime(&Stat_.st_atime), dst);
	return string(dst);
}

/* --------------------------------
 * t : C
 * -------------------------------- */

/* ŐV̏CԂ */
long DirectoryEntry::GetModifiedTime(void) const
{
	return Stat_.st_mtime;
}

/* ŐV̏CԂ */
void DirectoryEntry::GetModifiedTime(struct tm * dst) const
{
	struct tm*	tmp = localtime(&Stat_.st_mtime);
	*dst = *tmp;
}

/* Ct𐶐 */
void DirectoryEntry::GetModifiedDateString(char dst[]) const
{
	get_datestring(localtime(&Stat_.st_mtime), dst);
}

string DirectoryEntry::GetModifiedDateString(void) const
{
	char	dst[16];
	get_datestring(localtime(&Stat_.st_mtime), dst);
	return string(dst);
}

/* C𐶐 */
void DirectoryEntry::GetModifiedTimeString(char dst[]) const
{
	get_timestring(localtime(&Stat_.st_mtime), dst);
}

string DirectoryEntry::GetModifiedTimeString(void) const
{
	char	dst[16];
	get_timestring(localtime(&Stat_.st_mtime), dst);
	return string(dst);
}

/* --------------------------------
 * t : 쐬
 * -------------------------------- */

/* t@C쐬Ԃ */
long DirectoryEntry::GetCreatedTime(void) const
{
	return Stat_.st_ctime;
}

/* t@C쐬Ԃ */
void DirectoryEntry::GetCreatedTime(struct tm * dst) const
{
	struct tm*	tmp = localtime(&Stat_.st_ctime);
	*dst = *tmp;
}

/* 쐬t𐶐 */
void DirectoryEntry::GetCreatedDateString(char dst[]) const
{
	get_datestring(localtime(&Stat_.st_ctime), dst);
}

string DirectoryEntry::GetCreatedDateString(void) const
{
	char	dst[16];
	get_datestring(localtime(&Stat_.st_ctime), dst);
	return string(dst);
}

/* 쐬𐶐 */
void DirectoryEntry::GetCreatedTimeString(char dst[]) const
{
	get_timestring(localtime(&Stat_.st_ctime), dst);
}

string DirectoryEntry::GetCreatedTimeString(void) const
{
	char	dst[16];
	get_timestring(localtime(&Stat_.st_ctime), dst);
	return string(dst);
}

/* --------------------------------
 * tE
 * -------------------------------- */

/* t𐶐 */
void DirectoryEntry::get_datestring(struct tm* tmp, char dst[]) const
{
	sprintf(dst, "%04d/%02d/%02d", tmp->tm_year + 1900, tmp->tm_mon + 1, tmp->tm_mday);
}

/* 𐶐 */
void DirectoryEntry::get_timestring(struct tm* tmp, char dst[]) const
{
	sprintf(dst, "%02d:%02d:%02d", tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
}

/* --------------------------------
 * Agr[g
 * -------------------------------- */

/* Agr[g擾
 * <set>	Stat_
 */
bool DirectoryEntry::get_attribute(void)
{
	return GetAttribute(GetPathName().c_str(), &Stat_);
}

/* --------------------------------
 * q
 * -------------------------------- */

/* q(t@CETutH_)Ԃ */
DirectoryEntry & DirectoryEntry::GetChild(int index)
{
	int	idx = 0;;
	for (DirectoryEntryCollection::iterator itr=SubEntries_.begin(); itr!=SubEntries_.end(); ++itr) {
		if (idx == index) {
			DirectoryEntry*	entry = dynamic_cast<DirectoryEntry*>(itr->second);
			return *entry;
		}
		++idx;
	}
	char	str[30];
	sprintf(str, "%d", index);
	throw IllegalParameterException("DirectoryEntry::GetChild(int)", "index", str, 0);
}

DirectoryEntry & DirectoryEntry::GetChild(std::string Name)
{
	return *dynamic_cast<DirectoryEntry*>(SubEntries_.find(Name)->second);
}

/* qS폜
 * <set>	SubEntries_
 *		Searched_
 */
void DirectoryEntry::delete_children(void)
{
	for (DirectoryEntryCollection::iterator itr=SubEntries_.begin(); itr!=SubEntries_.end(); ++itr)
		delete (itr->second);
	SubEntries_.clear();
	Searched_ = false;
}

/* oRs[
 * <set>	Name_
 *		Dir_
 * 		Stat_
 * 		SubEntries_
 *		Searched_
 */
void DirectoryEntry::copy(const DirectoryEntry & other)
{
	delete_children();
	
	Name_ = other.Name_;
	Dir_ = other.Dir_;
	
	Stat_ = other.Stat_;
	
	for (DirectoryEntryCollection::const_iterator itr=other.SubEntries_.begin(); itr!=other.SubEntries_.end(); ++itr)
		SubEntries_.insert(DirectoryEntryCollection::value_type(itr->first, &itr->second->Clone()));
	
	Searched_ = other.Searched_;
}

/* ----------------------------------------------------------------
 * Operator Overloadings
 * ---------------------------------------------------------------- */

DirectoryEntry & DirectoryEntry::operator=(const DirectoryEntry & other)
{
	copy(other);
}

} /* namespace DirectoryEntryPackage */
/*-------1---------2---------3---------4---------5---------6---------7---------8---------9--------*/
/*  end of text */
