/*
 * DirectoryEntry
 * 	version 1.0.0
 */

#ifndef DIRECTORY_ENTRY_H
#define DIRECTORY_ENTRY_H

#include <time.h>
#include <sys/stat.h>
#include <string>
#include <map>
#include "UpdateException.hpp"

#define	DIRSTR_CURRENT	"."
#define	DIRSTR_PARENT	".."

namespace DirectoryEntryPackage {

class	DirectoryEntry;
class	Directory;
class	File;

typedef	std::map<std::string, DirectoryEntry*>	DirectoryEntryCollection;

/* ================================================================
 * O
 * ================================================================ */

class DirectoryEntryException : public UpdateException {
public:
	DirectoryEntryException(const std::string m, const std::string p, const std::string v, int n)
		: UpdateException(m, p, n), Value(v) {}
	~DirectoryEntryException() throw() {}
	std::string	Value;
};

/* O:NX */
class ClassInitializeException : public DirectoryEntryException {
public:
	ClassInitializeException(const std::string m, const std::string p, const std::string v, int n)
		: DirectoryEntryException(m, p, v, n) {}
	~ClassInitializeException() throw() {}
};

/* O:w */
class ParameterNOTSpecifiedException : public DirectoryEntryException {
public:
	ParameterNOTSpecifiedException(const std::string m, const std::string p, const std::string v, int n)
		: DirectoryEntryException(m, p, v, n) {}
	~ParameterNOTSpecifiedException() throw() {}
};

/* O:s */
class IllegalParameterException : public DirectoryEntryException {
public:
	IllegalParameterException(const std::string m, const std::string p, const std::string v, int n)
		: DirectoryEntryException(m, p, v, n) {}
	~IllegalParameterException() throw() {}
};

/* O:fBNg/t@C݂Ȃ */
class EntryNOTFoundException : public DirectoryEntryException {
public:
	EntryNOTFoundException(const std::string m, const std::string p, const std::string v, int n)
		: DirectoryEntryException(m, p, v, n) {}
	~EntryNOTFoundException() throw() {}
};

/* ================================================================
 * DirectoryEntryNX
 * ================================================================ */

class DirectoryEntry {
public:
	/* fBNg(Dir)ăt@CEtH_𒊏o
	 * <param>	Dir		fBNg
	 * 		Entries		i[p
	 *		recursive	true:TuEfBNgċAIɒo
	 * <exception>	EntryNOTFoundException
	 *		ParameterNOTSpecifiedException
	 *		IllegalParameterException
	 * <set>	Entries
	 */
	static void SearchChildren(const char * Dir, DirectoryEntryCollection & Entries, bool recursive=false);
	/* DirectoryEntryIuWFNg𐶐 */
	static DirectoryEntry* Create(const char * Dir, const char * Name);
	/* t@C̑擾 */
	static bool GetAttribute(const char * Path, struct stat* Stat);
	/* PATH(O+ꏊ)Ԃ */
	static std::string GetPathName(const std::string & Dir, const std::string & Name);
/*
 * Constructors / Destructors
 */
public:
	/* Constructor
	 * <param>	Dir
	 * 		Name
	 * <exception>	IllegalParameterException
	 *		ClassInitializeException
	 */
	DirectoryEntry(const char * Dir, const char * Name);
	/* Destructor */
	virtual ~DirectoryEntry() { delete_children(); }
protected:
	/* Default Constructor */
	DirectoryEntry() : SubEntries_() {}
	/* Copy Constructor */
	DirectoryEntry(const DirectoryEntry & other) { copy(other); }
/*
 * Methods
 */
public:
	/* OԂ */
	std::string GetName(void) const { return Name_; }
	/* fBNg(ꏊ)Ԃ */
	std::string GetDir(void) const { return Dir_; }
	/* PATH(O+ꏊ)Ԃ */
	std::string GetPathName(void) const;
	std::string GetFullPathName(void) const;
	
	/* gq */
	bool HasExtension(const std::string & ext) const;
	
	/* ŐṼANZXԂ */
	long GetAccessedTime(void) const;
	void GetAccessedTime(struct tm*) const;
	void GetAccessedDateString(char*) const;
	void GetAccessedTimeString(char*) const;
	std::string GetAccessedDateString() const;
	std::string GetAccessedTimeString() const;
	/* ŐV̏CԂ */
	long GetModifiedTime(void) const;
	void GetModifiedTime(struct tm*) const;
	void GetModifiedDateString(char*) const;
	void GetModifiedTimeString(char*) const;
	std::string GetModifiedDateString() const;
	std::string GetModifiedTimeString() const;
	/* t@C쐬Ԃ */
	long GetCreatedTime(void) const;
	void GetCreatedTime(struct tm*) const;
	void GetCreatedDateString(char*) const;
	void GetCreatedTimeString(char*) const;
	std::string GetCreatedDateString() const;
	std::string GetCreatedTimeString() const;
	
	/* q(t@CETutH_)Ԃ */
	DirectoryEntry & GetChild(int index);
	DirectoryEntry & GetChild(std::string);
	/* q(t@CETutH_)̐Ԃ */
	int GetChildrenCount(void) const { return SubEntries_.size(); }
	/* qĂ邩ǂԂ
	 * <ret>	true:qĂȂ(File)
	 *		false:qĂ(Directory)
	 */
	virtual bool IsTerminalNode(void) const = 0;
	/* q
	 * <set>	SubEntries_
	 *		Searched_
	 * <exception>	std::runtime_error	qĂȂ(File, IsTerminalNode()trueԂ)
	 */
	virtual void SearchChildren(bool recursive=false) = 0;
	/* g̕𐶐ĕԂ */
	virtual DirectoryEntry & Clone(void) const = 0;
protected:
	/* t𐶐 */
	void get_datestring(struct tm*, char*) const;
	/* 𐶐 */
	void get_timestring(struct tm*, char*) const;
	/* Agr[g擾
	 * <set>	Stat_
	 */
	bool get_attribute(void);
	/* qS폜
	 * <set>	SubEntries_
	 *		Searched_
	 */
	void delete_children(void);
	/* oRs[
	 * <set>	Name_
	 *		Dir_
	 * 		Stat_
	 * 		SubEntries_
	 *		Searched_
	 */
	virtual void copy(const DirectoryEntry &);
/*
 * Operator Overloadings
 */
public:
	virtual DirectoryEntry & operator=(const DirectoryEntry &);
/*
 * Fields
 */
protected:
	struct stat	Stat_;
	std::string	Name_;		/* O */
	std::string	Dir_;		/* fBNg(ꏊ) */
	
	bool	Searched_;
	DirectoryEntryCollection	SubEntries_;
}; /* end of class DirectoryEntry */

/* ================================================================
 * DirectoryNX
 * ================================================================ */
class Directory : public DirectoryEntry {
/*
 * Constructors / Destructors
 */
public:
	/* Constructor
	 * <param>	Dir
	 * 		Name
	 * <exception>	EntryNOTFoundException
	 *		ParameterNOTSpecifiedException
	 *		IllegalParameterException
	 */
	Directory(const char * Dir, const char * Name);
	~Directory() {}
private:
	Directory() {}
/*
 * Methods
 */
public:
	bool IsTerminalNode(void) const { return false; }
	/* q */
	void SearchChildren(bool recursive=false);
	/* g̕𐶐ĕԂ */
	DirectoryEntry & Clone(void) const;
}; /* end of class Directory */

/* ================================================================
 * FileNX
 * ================================================================ */
class File : public DirectoryEntry {
/*
 * Constructors / Destructors
 */
public:
	/* Constructor
	 * <param>	Dir
	 * 		Name
	 * <exception>	EntryNOTFoundException
	 *		ParameterNOTSpecifiedException
	 *		IllegalParameterException
	 */
	File(const char * Dir, const char * Name);
	~File() {}
private:
	File() {}
/*
 * Methods
 */
	bool IsTerminalNode(void) const { return true; }
	void SearchChildren(bool recursive=false) { throw std::runtime_error("File::SearchChildren() NOT implemented."); }
	/* g̕𐶐ĕԂ */
	DirectoryEntry & Clone(void) const;
	
	/* t@C̃TCYԂ */
	long GetSize(void) const { return Stat_.st_size; }
}; /* end of class File */

} /* namespace DirectoryEntryPackage */
#endif	/* #ifndef DIRECTORY_ENTRY_H */
/*-------1---------2---------3---------4---------5---------6---------7---------8---------9--------*/
/*  end of text */
