/*
 * UpdateCommand
 * 	version 1.0.0
 */
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/cygwin.h>
#include <errno.h>
#include <algorithm>
#include <cstdio>
#include <algorithm>
#include <iostream>
#include <stdexcept>

#include "DirectoryEntry.hpp"
#include "UpdateParams.hpp"
#include "UpdateCommand.hpp"

using namespace std;
using namespace DirectoryEntryPackage;

// #define	DEBUG

/* ================================================================
 * UpdateCommandNX
 * ================================================================ */

UpdateCommand::~UpdateCommand()
{
	if (Entries1 != NULL) {
		delete_entries(*Entries1);
		delete Entries1;
	}
	if (Entries2 != NULL) {
		delete_entries(*Entries2);
		delete Entries2;
	}
}

/* ㏈ */
void UpdateCommand::delete_entries(DirectoryEntryPackage::DirectoryEntryCollection & entries)
{
	for (DirectoryEntryCollection::iterator itr=entries.begin(); itr!=entries.end(); ++itr)
		delete itr->second;
	entries.clear();
}

/* ----------------------------------------------------------------
 * s
 * ---------------------------------------------------------------- */

/* --------------------------------
 * \
 * -------------------------------- */

/*
 * o:
 *	DIR1=****************************
 *	DIR2=****************************
 *	EXT=.xxx .yyy .zzz
 *	D/F DIR1                   DIR2                 FileName
 *	---+-------------------+--+--------------------+-----------------------------
 *	F   YYYY/MM/DD hh:mm:ss -> YYYY/MM/DD hh:mm:ss  ****************************
 *	F   YYYY/MM/DD hh:mm:ss <- YYYY/MM/DD hh:mm:ss  ****************************
 *	F   YYYY/MM/DD hh:mm:ss    YYYY/MM/DD hh:mm:ss  ****************************
 *	F   YYYY/MM/DD hh:mm:ss    YYYY/MM/DD hh:mm:ss  ****************************
 *	F   YYYY/MM/DD hh:mm:ss    YYYY/MM/DD hh:mm:ss  ****************************
 */

// #define	SUPPORT_SUBDIR

#if defined(SUPPORT_SUBDIR)
	static	string	DF("F  ");
#endif
static	string	TIMEBLANK("                   ");
static	string	SP(" ");
static	string	TAB4("    ");
static	string	LARROW("<-");
static	string	RARROW("->");
static	string	ACTBLANK("  ");
static	string	COPIED(" *");
static	string	COPYFAILED(" -");

void UpdateCommand::put_dirs(const UpdateParams & params) const
{
	cout << "DIR1=" << params.Dir1 << endl;
	cout << "DIR2=" << params.Dir2 << endl;
	if (params.Debug) {
		char	fullpath[260];
		cygwin_conv_to_full_posix_path(params.Dir1.c_str(), fullpath);
		cout << TAB4 << fullpath << endl;
		cygwin_conv_to_full_posix_path(params.Dir2.c_str(), fullpath);
		cout << TAB4 << fullpath << endl;
	}
}

void UpdateCommand::put_ext(const UpdateParams & params) const
{
	if (params.Extensions.size() == 0)
		return;
	cout << "EXT=";
	for (ExtCollection::const_iterator itr=params.Extensions.begin(); itr!=params.Extensions.end(); ++itr)
		cout << *itr << " ";
	cout << endl;
}

void UpdateCommand::put_header(void) const
{
#if defined(SUPPORT_SUBDIR)
	cout << "D/F DIR1                   DIR2                FileName (*=copied)" << endl;
	cout << "---+-------------------+--+-------------------+-----------------------------" << endl;
#else
	cout << "DIR1                   DIR2                FileName (*=copied)" << endl;
	cout << "-------------------+--+-------------------+-----------------------------" << endl;
#endif
}

string UpdateCommand::get_datetimestring(const DirectoryEntry * entry) const
{
	string	str;
	
	if (entry)
		str = entry->GetModifiedDateString() + SP + entry->GetModifiedTimeString();
	else	str = TIMEBLANK;
	return str;
}

void UpdateCommand::put_status(const UpdateParams & params, const DirectoryEntry * srcentry, const DirectoryEntry * dstentry, bool reverse, ACTION action) const
{
	string	DIR1;
	string	ACT;
	string	DIR2;
	string	NAME;
	
	if (reverse) {
		DIR1 = get_datetimestring(dstentry);
		DIR2 = get_datetimestring(srcentry);
	} else {
		DIR1 = get_datetimestring(srcentry);
		DIR2 = get_datetimestring(dstentry);
	}
	switch (action) {
	case ADD:
	case OVERWRITE:
		ACT = RARROW;
		break;
	case ADD_R:
	case OVERWRITE_R:
		ACT = LARROW;
		break;
	case NOTHING:
	default:
		ACT = ACTBLANK;
		break;
	}
	NAME = srcentry->GetName();
	if (params.Debug && params.DoExec)
		cout << "--------" << endl;
#if defined(SUPPORT_SUBDIR)
	cout << DF << SP;
#endif
	cout << DIR1 << SP << ACT << SP << DIR2 << SP << NAME;
	/* ł͉sȂ */
	if (params.Debug && params.DoExec)
		cout << endl;
}

/* --------------------------------
 * f
 * -------------------------------- */

/* t@C̃^CX^vr */
bool UpdateCommand::check_datetime(const DirectoryEntry & srcentry, const DirectoryEntry & dstentry, bool reverse) const
{
	long	datetime1 = srcentry.GetModifiedTime();
	long	datetime2 = dstentry.GetModifiedTime();
	if (!reverse)
		return (datetime1 > datetime2);
	else	return (datetime1 < datetime2);
}

UpdateCommand::ACTION UpdateCommand::get_action(const UpdateParams & params, const DirectoryEntry * srcentry, const DirectoryEntry * dstentry, bool reverse) const
{
	ACTION	action;
	
	if (params.Direction != UpdateParams::BOTH) {
		/* ʏ(->) */
		if (dstentry != NULL) {
			if (params.Overwrite == UpdateParams::NOTEXISTS)
				action = SKIP;
			else if (check_datetime(*srcentry, *dstentry, reverse))
				action = OVERWRITE;	/* -> */
			else	action = NOTHING;
		} else {
			if (params.Overwrite == UpdateParams::NOTEXISTS)
				action = ADD;		/* -> */
			else if (params.Overwrite == UpdateParams::EXISTS)
				action = NOTHING;
			else	action = ADD;		/* -> */
		}
	} else {
		/* o */
		if (!reverse) {
			/* -> */
			if (dstentry != NULL) {
				if (params.Overwrite == UpdateParams::NOTEXISTS)
					action = NOTHING;
				else if (check_datetime(*srcentry, *dstentry, false))
					action = OVERWRITE;	/* -> */
				else if (check_datetime(*srcentry, *dstentry, true))
					action = OVERWRITE_R;	/* <- */
				else	action = NOTHING;
			} else {
				if (params.Overwrite == UpdateParams::NOTEXISTS)
					action = ADD;	/* -> */
				else if (params.Overwrite == UpdateParams::EXISTS)
					action = NOTHING;
				else	action = ADD;	/* -> */
			}
		} else {
			/* <- */
			if (dstentry != NULL) {
				action = SKIP;
				//throw invalid_argument(dstentry->GetPathName());
			} else {
				if (params.Overwrite == UpdateParams::NOTEXISTS || params.Overwrite == UpdateParams::ADD)
					action = ADD_R;	/* <- */
				else	action = NOTHING;
			}
		}
	}
	if (action == NOTHING) {
		if (params.Overwrite == UpdateParams::NOTEXISTS || params.SkipView)
			action = SKIP;
	}
	return action;
}

/* --------------------------------
 * s Main
 * -------------------------------- */

bool UpdateCommand::set_modtime(const UpdateParams & params, const std::string & dstpath, long time) const
{
	struct timeval	times[2];
	times[0].tv_sec = time;
	times[0].tv_usec = 0;
	times[1].tv_sec = time;
	times[1].tv_usec = 0;
	if (utimes(dstpath.c_str(), times) == -1)
		throw UpdateCommandException("set_modtime(const UpdateParams&,const string&,long)", dstpath, "utimes()Ɏs܂", errno);
	return true;
}

bool UpdateCommand::copy(const UpdateParams & params, const std::string & srcpath, const std::string & dstpath, long mtime) const
{
	char	srcfullpath[260];
	char	dstfullpath[260];
	cygwin_conv_to_full_posix_path(srcpath.c_str(), srcfullpath);
	cygwin_conv_to_full_posix_path(dstpath.c_str(), dstfullpath);
	if (params.Debug)
		return true;
	
	struct stat	st_buf;
	
	/* srcJăɓǂݍ */
	int	from = open(srcfullpath, O_RDONLY, 0644);	/* 0644='-wr--r--r-',O_RDONLYȂ̂Ŗ */
	if (from == -1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", srcfullpath, "t@CI[vł܂", errno);
	if (fstat(from, &st_buf) == -1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", srcfullpath, "Xe[^X擾ł܂", errno);
	size_t	size = st_buf.st_size;
	void *	from_mmap = mmap(NULL, size, PROT_READ, MAP_PRIVATE, from, 0);
	if (from_mmap == (void*)-1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", srcfullpath, "Ƀ}bvł܂", errno);
	
	/* dstJ */
	int	to = open(dstfullpath, O_CREAT|O_RDWR, st_buf.st_mode);
	if (to == -1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", dstfullpath, "t@CI[vł܂", errno);
	
	/* copy */
	char *	p = (char *)from_mmap;
	const char * const	endp = (const char *)from_mmap + size;
	while (p < endp) {
		int	num_bytes = write(to, p, endp - p);
		p += num_bytes;
	}
	// assert(p == endp);
	fsync(to);
	
	/* Еt:̊JƊJt@C̃N[Y */
	if (munmap(from_mmap, size) == -1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", srcfullpath, "munmap()Ɏs܂", errno);
	if (close(to) == -1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", dstfullpath, "close()Ɏs܂", errno);
	if (close(from) == -1)
		throw UpdateCommandException("copy(const UpdateParams&,const string&,const string&)", srcfullpath, "close()Ɏs܂", errno);
	
	return set_modtime(params, dstfullpath, mtime);;
}

bool UpdateCommand::copy(const UpdateParams & params, const DirectoryEntry * srcentry, const DirectoryEntry * dstentry, const std::string & dstdir) const
{
	string	dstpath;
	if (dstentry)
		dstpath = dstentry->GetPathName();
	else	dstpath = dstdir + "\\" + srcentry->GetName();
	if (params.Debug) {
		cout << TAB4 << "(SRC=" << srcentry->GetFullPathName() << endl;
		cout << TAB4 << TAB4 << "MTIME=" << srcentry->GetModifiedTime() << endl;
		if (dstentry) {
			cout << TAB4 << "(DST=" << dstentry->GetFullPathName() << endl;
			cout << TAB4 << TAB4 << "MTIME=" << dstentry->GetModifiedTime() << endl;
		} else {
			char	fullpath[260];
			cygwin_conv_to_full_posix_path(dstpath.c_str(), fullpath);
			cout << TAB4 << "(DST=" << fullpath << endl;
		}
	}
	bool	result = copy(params, srcentry->GetPathName(), dstpath, srcentry->GetModifiedTime());
	
	if (params.Debug)
		cout << TAB4;
	if (result)
		cout << COPIED;
	else	cout << COPYFAILED;
	return result;
}

bool UpdateCommand::copy(const UpdateParams & params, const DirectoryEntry * srcentry, const DirectoryEntry * dstentry, bool reverse, ACTION action) const
{
	bool	result;
	switch (action) {
	case ADD:
	case OVERWRITE:
		/* -> */
		if (!reverse)
			result = copy(params, srcentry, dstentry, params.Dir2);
		else	result = copy(params, dstentry, srcentry, params.Dir2);
		break;
	case ADD_R:
	case OVERWRITE_R:
		/* <- */
		if (!reverse)
			result = copy(params, dstentry, srcentry, params.Dir1);
		else	result = copy(params, srcentry, dstentry, params.Dir1);
		break;
	case NOTHING:
	default:
		/* DO NOTHING */
		result = true;
		break;
	}
	return result;
}

bool UpdateCommand::hasextension(const UpdateParams & params, const DirectoryEntryPackage::DirectoryEntry * entry) const
{
	if (params.Extensions.size() == 0)
		return true;
	for (ExtCollection::const_iterator itr=params.Extensions.begin(); itr!=params.Extensions.end(); ++itr) {
		if (entry->HasExtension(*itr))
			return true;
	}
	return false;
}

bool UpdateCommand::exec(const UpdateParams & params, const DirectoryEntryCollection & entries1, const DirectoryEntryCollection & entries2, bool reverse) const
{
	bool	result = true;
	for (DirectoryEntryCollection::const_iterator itr=entries1.begin(); itr!=entries1.end(); ++itr) {
		const DirectoryEntry *	srcentry = itr->second;
		if (srcentry->IsTerminalNode()) {	/* t@Cr */
			if (hasextension(params, srcentry)) {
				string	srcname = srcentry->GetName();
				/* t@CɕϊĒo */
				std::transform(srcname.begin(), srcname.end(), srcname.begin(), (int (*)(int))std::tolower);
				DirectoryEntryCollection::const_iterator	dstitr = entries2.find(srcname);
				const DirectoryEntry *	dstentry;
				if (dstitr != entries2.end())
					dstentry = dstitr->second;
				else	dstentry = NULL;
				ACTION	action = get_action(params, srcentry, dstentry, reverse);
				if (action != SKIP) {
					put_status(params, srcentry, dstentry, reverse, action);
					if (params.DoExec) {
						result = copy(params, srcentry, dstentry, reverse, action);
						if (!result)
							break;
					}
					cout << endl;	/* put_status()ŕ\s̉s */
				}
			}
		}
	}
	return result;
}

void UpdateCommand::Execute(UpdateParams & params)
{
	if (params.ShowHelp) {
		params.Usage();
		return;
	}
	switch (params.ErrorStatus()) {
	case UpdateParams::MISSING:
		cout << "ERROR:(Dir1 == '' || Dir2 == '')" << endl;
		params.Usage();
		return;
	case UpdateParams::EQUALS:
		cout << "ERROR:(DIR1 == DIR2)" << endl;
		params.Usage();
		return;
	case UpdateParams::ERROR:
		cout << "ERROR" << endl;
		params.Usage();
		return;
	case UpdateParams::SHOWHELP:
		params.Usage();
		return;
	}
	
	if (!Entries1) {
		Entries1 = new DirectoryEntryCollection;
		DirectoryEntry::SearchChildren(params.Dir1.c_str(), *Entries1);
	}
	if (!Entries2) {
		Entries2 = new DirectoryEntryCollection;
		DirectoryEntry::SearchChildren(params.Dir2.c_str(), *Entries2);
	}
	
	put_dirs(params);
	put_ext(params);
	put_header();
	if (exec(params, *Entries1, *Entries2, false)) {	/* 12 */
		if (params.Direction == UpdateParams::BOTH && params.Overwrite != UpdateParams::EXISTS) {
#if defined(DEBUG)
			cout << "<REVERSE>" << endl;
#endif
			exec(params, *Entries2, *Entries1, true);	/* 12 */
		}
	}
}

/*-------1---------2---------3---------4---------5---------6---------7---------8---------9--------*/
/*  end of text */
