/*
 * UpdateCommand
 * 	version 1.0.0
 */
#ifndef UPDATECOMMAND_H
#define UPDATECOMMAND_H

#include <string>
#include "DirectoryEntry.hpp"
#include "UpdateParams.hpp"
#include "UpdateException.hpp"

/* ================================================================
 * O
 * ================================================================ */

class UpdateCommandException : public UpdateException {
public:
	UpdateCommandException(const std::string method, const std::string param, const std::string msg, int errnum)
		: UpdateException(method, param, errnum), Msg(msg) {}
	~UpdateCommandException() throw() {}
	std::string	Msg;
};

/* ================================================================
 * UpdateCommandNX
 * ================================================================ */

class UpdateCommand {
public:
	UpdateCommand() : Entries1(NULL), Entries2(NULL) {}
	~UpdateCommand();
	void Execute(UpdateParams &);
private:
	typedef enum {
		CHECK, ADD, OVERWRITE, ADD_R, OVERWRITE_R, NOTHING, SKIP
	} ACTION;
	/* \ */
	void put_dirs(const UpdateParams &) const;
	void put_ext(const UpdateParams & params) const;
	void put_header(void) const;
	std::string get_datetimestring(const DirectoryEntryPackage::DirectoryEntry *) const;
	void put_status(const UpdateParams &, const DirectoryEntryPackage::DirectoryEntry *, const DirectoryEntryPackage::DirectoryEntry *, bool, ACTION) const;
	/* f */
	bool check_datetime(const DirectoryEntryPackage::DirectoryEntry &, const DirectoryEntryPackage::DirectoryEntry &, bool) const;
	ACTION get_action(const UpdateParams &, const DirectoryEntryPackage::DirectoryEntry *, const DirectoryEntryPackage::DirectoryEntry *, bool) const;
	/* s */
	bool exec(const UpdateParams &, const DirectoryEntryPackage::DirectoryEntryCollection &, const DirectoryEntryPackage::DirectoryEntryCollection &, bool) const;
	bool hasextension(const UpdateParams &, const DirectoryEntryPackage::DirectoryEntry *) const;
	bool copy(const UpdateParams &, const DirectoryEntryPackage::DirectoryEntry *, const DirectoryEntryPackage::DirectoryEntry *, bool, ACTION) const;
	bool copy(const UpdateParams &, const DirectoryEntryPackage::DirectoryEntry *, const DirectoryEntryPackage::DirectoryEntry *, const std::string &) const;
	bool copy(const UpdateParams &, const std::string &, const std::string &, long) const;
	bool set_modtime(const UpdateParams &, const std::string &, long) const;
	/* ㏈ */
	void delete_entries(DirectoryEntryPackage::DirectoryEntryCollection&);
private:
	DirectoryEntryPackage::DirectoryEntryCollection*	Entries1;
	DirectoryEntryPackage::DirectoryEntryCollection*	Entries2;
}; /* end of class UpdateCommand */

#endif	/* #ifndef UPDATECOMMAND_H */
/*-------1---------2---------3---------4---------5---------6---------7---------8---------9--------*/
/*  end of text */
