/*
 * UpdateParams
 * 	version 1.0.0
 */
#include <sys/cygwin.h>
#include <cstring>
#include <string>
#include <stdexcept>
#include <iostream>
#include <fstream>

#include "UpdateParams.hpp"

using namespace std;

#define	OPT_PREFIX_CHAR			'-'
#define	OPT_PREFIX			"-"

#define	OPT_HELP_SUFFIX			"h"
#define	OPT_HELP_Q_SUFFIX		"?"
#define	OPT_HELP_FULL_SUFFIX		"help"
#define	OPT_VERSION_SUFFIX		"v"
#define	OPT_VERSION_FULL_SUFFIX		"version"
#define	OPT_HELP			OPT_PREFIX OPT_HELP_SUFFIX
#define	OPT_HELP_Q			OPT_PREFIX OPT_HELP_Q_SUFFIX
#define	OPT_HELP_FULL			OPT_PREFIX OPT_PREFIX OPT_HELP_FULL_SUFFIX
#define	OPT_VERSION			OPT_PREFIX OPT_VERSION_SUFFIX
#define	OPT_VERSION_FULL		OPT_PREFIX OPT_PREFIX OPT_VERSION_FULL_SUFFIX

#define	OPT_EXEC_SUFFIX			"x"
#define	OPT_DIRECT_BOTH_SUFFIX		"b"
#define	OPT_DIRECT_REV_SUFFIX		"r"
#define	OPT_OVWR_ADD_SUFFIX		"a"
#define	OPT_OVWR_EXISTS_SUFFIX		"e"
#define	OPT_OVWR_NOTEXISTS_SUFFIX	"n"
#define	OPT_SKIPVIEW_SUFFIX		"s"
#define	OPT_EXT_SUFFIX			"k"
#define	OPT_FILE_SUFFIX			"f"
#define	OPT_EXEC			OPT_PREFIX OPT_EXEC_SUFFIX
#define	OPT_DIRECT_BOTH			OPT_PREFIX OPT_DIRECT_BOTH_SUFFIX
#define	OPT_DIRECT_REV			OPT_PREFIX OPT_DIRECT_REV_SUFFIX
#define	OPT_OVWR_ADD			OPT_PREFIX OPT_OVWR_ADD_SUFFIX
#define	OPT_OVWR_EXISTS			OPT_PREFIX OPT_OVWR_EXISTS_SUFFIX
#define	OPT_OVWR_NOTEXISTS		OPT_PREFIX OPT_OVWR_NOTEXISTS_SUFFIX
#define	OPT_SKIPVIEW			OPT_PREFIX OPT_SKIPVIEW_SUFFIX

#define	OPT_EXT				OPT_PREFIX OPT_EXT_SUFFIX
#define	OPT_EXT_SUFFIX_CHAR		'k'
#define	EXT_PREFIX_CHAR			'.'
#define	EXT_PREFIX			"."

#define	OPT_FILE			OPT_PREFIX OPT_FILE_SUFFIX
#define	OPT_FILE_SUFFIX_CHAR		'f'

#define	OPT_EXEC_SUFFIX_CHAR		'x'

#define	OPT_DEBUG_SUFFIX		"D"
#define	OPT_DEBUG			OPT_PREFIX OPT_DEBUG_SUFFIX
#define	OPT_DEBUG_FULL_SUFFIX		"DEBUG"
#define	OPT_DEBUG_FULL			OPT_PREFIX OPT_PREFIX OPT_DEBUG_FULL_SUFFIX

/* ================================================================
 * UpdateParamsNX
 * ================================================================ */

/* gp@\ */
void UpdateParams::Usage(void) const
{
	cout << "USAGE>update ";
	cout << "[";
		cout << OPT_PREFIX;
		cout << OPT_OVWR_ADD_SUFFIX OPT_DIRECT_BOTH_SUFFIX OPT_OVWR_EXISTS_SUFFIX OPT_OVWR_NOTEXISTS_SUFFIX OPT_DIRECT_REV_SUFFIX OPT_SKIPVIEW_SUFFIX OPT_EXEC_SUFFIX;
	cout << "]";
	cout << " <dir1> <dir2>";
	cout << " [";
		cout << OPT_EXT "<.ext> ...";
	cout << "]";
	cout << " [";
		cout << OPT_FILE "<filepath> ...";
	cout << "]";
	cout << endl;
	cout << "Options:" OPT_EXEC "  XVs" << endl;
	cout << "        " OPT_DIRECT_BOTH "  o@ɎQƂ" << endl;
	cout << "        " OPT_DIRECT_REV "  t@ɎQƂ" << endl;
	cout << "        " OPT_OVWR_ADD "  ݂Ȃt@CRs[" << endl;
	cout << "        " OPT_OVWR_EXISTS "  ݂t@Ĉ(*)" << endl;
	cout << "        " OPT_OVWR_NOTEXISTS "  ݂Ȃt@Ĉ" << endl;
	cout << "        " OPT_SKIPVIEW "  XVȂt@C͕\\Ȃ" << endl;
	cout << "        " OPT_EXT "  gqw肷" << endl;
	cout << "        " OPT_FILE "  XNvgt@Cw肷" << endl;
	cout << "    " OPT_DIRECT_REV_SUFFIX OPT_DIRECT_BOTH_SUFFIX "A" OPT_OVWR_ADD_SUFFIX OPT_OVWR_EXISTS_SUFFIX OPT_OVWR_NOTEXISTS_SUFFIX "͂ꂼrA̐ݒ肪㏑" << endl;

}

/* G[Xe[^XԂ */
UpdateParams::ERRORSTATUS UpdateParams::ErrorStatus(void) const
{
	if (Dir1 == "" && Dir2 == "")
		return SHOWHELP;
	if (Dir1 == "" || Dir2 == "")
		return MISSING;
	
	string	srcfullpath = fullpath(Dir1);
	string	dstfullpath = fullpath(Dir2);
	if (srcfullpath == dstfullpath)
		return EQUALS;
	
	return errorstatus;
}

/* POSIXtpXԂ */
std::string UpdateParams::fullpath(const std::string & path) const
{
	char	fullpath[260];
	cygwin_conv_to_full_posix_path(path.c_str(), fullpath);
	int	len = strlen(fullpath);
	if (fullpath[len - 1] != '/') {
		fullpath[len] = '/';
		fullpath[len + 1] = '\0';
	}
	return string(fullpath);
}

/* p[^Zbg */
void UpdateParams::set(std::string param)
{
	if (param.at(0) == OPT_PREFIX_CHAR) {
		if (param.length() == 1)
			throw invalid_argument(param);
		else if (param == OPT_EXEC)
			DoExec = true;
		else if (param == OPT_DIRECT_BOTH)
			Direction = BOTH;
		else if (param == OPT_DIRECT_REV)
			Direction = REV;
		else if (param == OPT_OVWR_ADD)
			Overwrite = ADD;
		else if (param == OPT_OVWR_EXISTS)
			Overwrite = EXISTS;
		else if (param == OPT_OVWR_NOTEXISTS)
			Overwrite = NOTEXISTS;
		else if (param == OPT_SKIPVIEW)
			SkipView = true;
		else if (param == OPT_HELP || param == OPT_HELP_Q || param == OPT_HELP_FULL)
			ShowHelp = true;
		else if (param == OPT_VERSION || param == OPT_VERSION_FULL)
			ShowVersion = true;
		else if (param == OPT_DEBUG || param == OPT_DEBUG_FULL)
			Debug = true;
		else if (param.at(1) == OPT_EXT_SUFFIX_CHAR) {
			if (param.length() > 2)
				Extensions.insert(param.substr(2));
			else	throw invalid_argument(param);
		} else if (param.at(1) == OPT_FILE_SUFFIX_CHAR) {
			if (param.length() > 2)
				Scripts.push_back(param.substr(2));
			else	throw invalid_argument(param);
		} else	throw invalid_argument(param);
	} else {
		if (Dir1 == "")
			Dir1 = param;
		else if (Dir2 == "")
			Dir2 = param;
		else	throw invalid_argument(param);
	}
}

/* XNvgt@Cp[^ǂݍ
 *	'-f''-x'͏Ȃ
 */
void UpdateParams::load_script(const std::string & path)
{
	/* open */
	ifstream	ifs(path.c_str());
	if (ifs.fail()) {
		cout << "G[:XNvgt@C'" << path << "'J܂" << endl;
		errorstatus = ERROR;
		return;
	}
	/* load & set */
	char	line[512];
	while (!ifs.eof()) {
		ifs.getline(line, sizeof(line));
		if (strlen(line) > 0) {
			/* '-f''-x'͏Ȃ */
			if (line[0] != OPT_PREFIX_CHAR || (line[1] != OPT_EXEC_SUFFIX_CHAR && line[1] != OPT_FILE_SUFFIX_CHAR)) {
				/* ̉s폜 */
				while (line[strlen(line) - 1] == '\n' || line[strlen(line) - 1] == '\r')
					line[strlen(line) - 1] = '\0';
				if (strlen(line) > 0)
					set(string(line));
			}
		}
	}
	/* close */
}

/* XNvgt@Cp[^ǂݍ */
void UpdateParams::load_scripts(void)
{
	for (ScriptCollection::const_iterator itr=Scripts.begin(); itr!=Scripts.end(); ++itr)
		load_script(*itr);
}

/* p[^Zbg */
void UpdateParams::Set(int argc, char * argv[])
{
	for (int index=1; index<argc; ++index)
		set(string(argv[index]));
	if (Scripts.size() > 0)
		load_scripts();
	if (Direction == REV)
		swap(Dir1, Dir2);
}

/*-------1---------2---------3---------4---------5---------6---------7---------8---------9--------*/
/*  end of text */
